# Working with files and directories

??? quote "First click and follow the instructions below only if you start the course at this stage! Otherwise skip this step!"
    {%
    include-markdown "pages/bash/bash-0-setup.md"
    %}

Let's be sure to be in the `script` directory before to continue the rest of the training:
```bash
cd ~/{{extra.working_directory}}/script
```

![](../images/3-navigation/tree_exo5.png){: style="height:250px"}

## Commands to keep in mind

| Action | Files | Folders |
| :-: | :-: | :-: |
| Inspect | ls | ls |
| View content | cat | ls |
| Navigate to |  | cd |
| Move | mv | mv |
| Copy | cp | cp -r |
| Create | nano | mkdir |
| Delete | rm | rmdir, rm -r |

## Creating directories

We have a `script` directory on one side and a `data` directory in another side. Let's suppose we will run analysis, it will be nice to have an `analysis` dedicated folder.

!!! question "How would you proceed?"

??? example "Click to show the solution"
    `mkdir analysis`

This is quite straightforward.

## Creating files

Let's now create a file... how would you proceed?  
Actually there are 2 main possibilities: 

* using a text editor
* redirecting a command output into a file using the special characters `>` or `>>`. But this is too early to learn this way for now.

There are plenty of text editors available: `ed`, `ex`, `vi`, `vim`, `emacs`, `nano`, etc.

As beginner, a good choice would be `nano`... why? Because the command how to use it are directly discribe in the screen it opens when you call it.

Let's create a file called `protocol.txt` using `nano`:

```bash
nano protocol.txt
```

You will end up in a window that represent the inside of your file, with the cursor ready to listen what you want to write. At the botton of the screen you will see description of few commands available. Indeed the mouse is useless here... everything has to pass by a command, which will be a combinaation of two characters where the first one is always `^` and corresponds to the **control key** on your keybord.

![](../images/4-creation/nano.png){: style="height:250px"}

Try to display the help `^G`. Not easy isnt'it? But it is like that, no way to make it easier... Depending the command line editor the command that can be executed and their key combinations differ. With nano what is conveniant it's that we do not need a command cheatsheets, the commands are described before our very eyes.

Let's now write someting: `I love guacamole`  
and save the result with `^O`

What happens? Actually nano before to quit and save ask you if you are happy with the file name where it will write. Yes sure! So to go ahead just type `↵ Return`. 

!!! question "Why `nano` ask if we wish to keep the file name?"

??? example "Click to show the solution"
    To be sure you don't fo any wrong manipulation and you are on the point to overwrite a file very important. Right, once the modification is done, there is no turning back.

You can now exit with `^X`.

!!! tip
    You can directly type `^X` when you are done with your edition. `nano` will ask first if you want to save the result, then ask if you are sure about the name of the file you are on the way to write in, and then exit.  


## Naming files and directories

There are few rules to respect to make your life smoother.

* No space
* No accent or special symbol! forbidden! `&` `~` `#` `”` `'` `{` `(` `[` `|` `\` `` ` `` `^` `@` `)` `]` `}` `$` `*` `%` `!` `/` `;` `,` `?`
* Use roman letters, numbers and `_` and `-`
* use file extension to guess what is inside `.fa` `.txt` `.bam` changing the extension of a fasta file by the `.bam` extension doesn’t somehow magically turn it into an alignment file. Files merely contain bytes; it’s up to us and our programs to interpret those bytes according. 
* Unix systems are case sensitive. Choose a way to write and follow it! `Sequence.fa` ≠ `SEQUENCE.fa` ≠ `sequence.fa`

## Moving files and directories

The `analysis` directory and the `protocol.txt` file are not located as suitable. We should move them more approriately.

The command `mv` allows to move file and directories. It follow this syntax:  
`mv <source> <target>`  
The difficult task here is to know the path relative/absolute of the source (what we want to move) and the path relative/absolute of the target (where we want to move it).   

Below is the current situation and the situation we want to end up:

![](../images/4-creation/move1.png){: style="height:250px"} ![](../images/4-creation/move2.png){: style="height:250px"}

!!! question "Move the `analysis` folder and `protocol.txt` accordingly."

??? example "Click to show the solution"
    using the absolute path:
    ```bash
    mv /home/<userName>/{{extra.working_directory}}/script/analysis /home/<userName>/{{extra.working_directory}}/analysis
    mv /home/<userName>/{{extra.working_directory}}/script/protocol.txt /home/<userName>/{{extra.working_directory}}/analysis/ 
    ```
    using the relative path:  
    ```bash
    mv analysis ..
    mv protocol.txt ../analysis/
    ```
    You could even mix abolute and relative path...

## Copy files and directories

Let's imagine you want your script `script1.sh` locally to run your analysis, but you want just a copy. You want to keep the original file as it is. 
The command to copy files and directories is `cp`. But for directories you will need to use the `-r` option that stands for recursive (copy also everything inside).

!!! question "Try to copy the `script1.sh` file inside `analysis` folder."

!!! tip 
    Once again you need to well understand where you are in your directory structure and use either the absolute or relative path of the file/folder you want to copy and the place you want to paste it.  
    Remmember that the current directory is also equivqlent to `.`

??? example "Click to show the solution"  
    ```bash
    cp ../script/script1.sh .
    ```

Let's practice a bit more and by trying to copy the entire `script` folder this time.

??? example "Click to show the solution"  
    ```bash
    cp -r ../script .
    ```
    /!\ If you use `cp -r ../script/ .` you will copy everything that is inside but not the `script` folder itself! So you see the difference?      
    Check the result:
    ```bash
    pwd
    ls
    ls script
    ```


### Your first globbing 

Before running you analysis you also need some data. Let's copy in the fastq files present in the `data` folder.
You could use the `cp` command for all the files one by one as we have seen previously.
But imagine you have hundred of those `fastq` files?  
Well Bash does carry out a very powerful feature, the `filename expansion`, a process known as globbing! (to not confound with regular expressions).

Here the list of globbing:

`?` Any single character
`*` Zero or more characters
`[]` Specify a range. Any character of the range or none of them by using ! inside the bracket.
`{term1,term2}` Specify a list of terms separated by commas and each term must be a name or a
wildcard.
`{term1..term2}` Called brace expansion, this syntax
expands all the terms between term1 and term2 (Letters or Integers).

For now you just has to rember the most important one, `*`.

Let's now copy the fastq files using the `*` filename expansion.

??? example "Click to show the solution"  
    ```bash
    cp ../data/*.fastq .
    ```
    Very convenient isn't it?

## Renaming files and directories

Sadly you realize your protocol is bad you will have to write another one. To keep a track of it you just want to rename it.
The easiest way to rename a file or a directory is to use the `mv` command.

!!! question "Change the `protocol.txt` file into `protocol_wrong.txt`."

??? example "Click to show the solution"  
    ```bash
    mv protocol.txt protocol_wrong.txt
    ```

If you were using a path, you just had to follow the same concept. Let's try by renaming the `protocol_wrong.txt` into `protocol_very_wrong.txt` using absolute path:

`mv /home/<userName>/{{extra.working_directory}}/script/protocol_wrong.txt /home/<userName>/{{extra.working_directory}}/script/protocol_very_wrong.txt`

!!! question "Let's finish this exercice by changing the name of the `script` directory into `script_copy`"
??? example "Click to show the solution"  
    ```bash
    mv script script_copy
    ls -Fa # to check everything went well
    ```

## Removing files and directories

Let's clean the mess we have done in this `analysis folder`. The command to remove files and directories is `rm`. In the case of directory you will have to use the `-r` option to make it recursive.

!!! question "remove the `script1.sh` file."

??? example "Click to show the solution"  
    ```bash
    rm script1.sh
    ```

!!! question "Now remove the `fastq` files using the filename expansion."

??? example "Click to show the solution"  
    ```bash
    rm *.fastq
    ```

!!! question "Let's finish this exercice by removing the `script_copy` directory."

??? example "Click to show the solution"  
    ```bash
    rm -r script_copy
    ls -Fa # to check everything went well
    ```

## linking files and directories

Actually in a real world what we have done so far was not very recommended (copy `fastq` files and `scripts.sh` file). 

!!! question "Do you guess why we should have avoid to copy those files?"

??? example "Click to show the solution"  
    Fastq files are generaly huge. Doing a copy double the disk space used by the file, because now you have it twice. It is recommended when possible to avoid to make copies of huge files.
    You may have continue to work on the `script1.sh` file located in the `script/` folder, but what if you forgot to copy the last version of it into your `analysis` directory?

In your new bioinformatician life you will often need to acces files far away many times (e.g. input files for a local script) and you will have to deal with long path. Long paths can be annoying...

There is a solution to ease your life, giving you the simplicity to have a file locally (no need to think about the path), to avoid having a copy of the file and to avoid dealing with long path: The **link**!

It exists two types of link: the `hard link` and the `soft link`.

### Hard link

A file is actually a reference to an existing inode (data structure on the file system that stores information).
Making a hard link to a file is essentially setting an additional reference to the same inode.
All hard links pointing to the same inode share the same set of data blocks on the disk.
If you delete one hard link, the data is not actually removed from disk until the last hard link is deleted.

![](../images/4-creation/hard_link.png){: style="height:350px"}

The hard link is made via the `ln` command.

```bash
ln existing_file [hard_link_name]
```

!!! Exercice
    Create a Hard link of the `data` folder
    ??? example "Click to show the solution"  
        ```ln ../data```
    What do you see?
    ??? example "Click to show the solution"  
        It is actually not possible to hard link a directory
    Run `ls -l` and then create a Hard link of the `data/data1.fa` file and run again `ls -l`
    ??? example "Click to show the solution"  
        ```bash
        ln ../data/data1.fa 
        ```
    create a Hard link of the `data/data1.fa` but with the choosen name `data1_hard.fa`  
    ??? example "Click to show the solution"  
        ```bash
        ln ../data/data1.fa data1_hard.fa
        ```
    Run again `ls -l`.  
    The second column is the number of hard links to the file. (For a directory, the number of hard links is the number of immediate subdirectories it has plus its parent directory and itself) We can see that we have currrently 3 "files" pointing to the same inode.


### Soft link

 The `soft link` is also known as `symbolic link`. It is a special sort of file that points at a different file. You could think of it like a shortcut.
 If you remove the original file, the link become inoperable. 

![](../images/4-creation/soft_link.png){: style="height:350px"}

 The solft link is made via the `ln -s` command.

```bash
ln -s existing_file [soft_link_name]
```

!!! Exercice
    Create a Soft link of the `script` folder
    ??? example "Click to show the solution"  
        ```ln  -s ../script```
    Run `ls -l`. What do you see?
    ??? example "Click to show the solution"  
        There is an extra arrow showing to what file my link point to.
    Create a Hard link of the `script/script1.sh` file and run again `ls -l`
    ??? example "Click to show the solution"  
        ```bash
        ln -s ../data/data1.fa
        ls -l
        ```

### Removing links

To remove links you can use `rm` like for files or the dedicated `unlink` command (the only difference is that unlink cannot remove directories).

!!! Exercice
    Remove all the link we set: `data`, `data1.fa`, `data1_hard.fa`, `script`, `script1.sh`

!!! Success "Quick recap"
    In this section we've learned:

    - How to create files and directories
    - How to move files and directories
    - How to rename files and directories
    - How to delete files and directories
    - How to link files and directories
